LAYER_DICT_NAMES = {
    "weights": "weights",
    "inp_size": "inp_size",
    "input_size": "inp_size",
    "bias": "biases",
    "biases": "biases",
    "shape": "shape",
    "units": "units",
    "layer_type": "layer_type",
    "dtype": "dtype",
    "activation": "activation",
    "decorator_params": "decorator_params",
}

VERSION = "1.1.4"

HEADER_OF_APG_FILE = f"""
Package: DEGANN
version: {VERSION}
Author: Alimov Pavel Gennadievich
"""
